#include <exec/types.h>
#include <exec/ports.h>
#include <exec/io.h>
#include <devices/ahi.h>

#include <proto/exec.h>

#include "sample.h"


void dprintf(const char *, ...);

#define D(x)    ;


struct Sample *CreateBeepSample(UWORD period, UWORD duration)
{
	struct Sample *beep;
	static const WORD sinewave[16] =
	{
		 0*256,  48*256,  86*256,  110*256,  120*256,  114*256,  86*256,  44*256,
		-1*256, -45*256, -84*256, -113*256, -124*256, -112*256, -84*256, -47*256
	};
	const int len_sinewave = sizeof(sinewave) / sizeof(sinewave[0]);
	ULONG micros, freq, samplesize;

	D(dprintf("CreateBeepSample: period %lu duration %lu\n", period, duration));

	micros = 16 * 400 * duration;

	/* min freq: 3546895/3134 = 1131
	 * max freq: 3546895/73   = 48587
	 */
	freq = 3546895 / period;

	samplesize = ((UQUAD) micros * freq) / 1000000 * sizeof(WORD);

	D(dprintf("CreateBeepSample: samplesize %lu\n", samplesize));

	beep = AllocMem(sizeof(*beep) + samplesize, MEMF_PUBLIC);
	if (beep)
	{
		ULONG numsines;
		float vol, volstep;
		WORD *pt, *end;
		int i;

		beep->SampleData = (APTR) (beep + 1);
		beep->SampleSize = samplesize;
		beep->UseCnt     = 1;

		pt = beep->SampleData;
		end = (APTR) (((char *) pt) + beep->SampleSize);

		numsines = samplesize / sizeof(sinewave);

		vol = 65536.0;
		volstep = vol / numsines;

		for (i = 0; i < numsines; i++)
		{
			int j;

			for (j = 0; j < len_sinewave; j++)
			{
				*pt++ = (LONG) (sinewave[j] * vol) >> 16;
			}
			vol -= volstep;
		}

		/* Fill rest with silence */
		while (pt < end)
		{
			*pt++ = 0;
		}
	}

	D(dprintf("CreateBeepSample: return 0x%lx\n", beep));

	return beep;
}

void ObtainSample(struct Sample *sample)
{
	D(dprintf("ObtainSample: sample 0x%lx\n", sample));

	if (sample)
	{
		sample->UseCnt++;

		D(dprintf("ObtainSample: sample UseCnt %ld\n", sample->UseCnt));
	}
}

void ReleaseSample(struct Sample *sample)
{
	D(dprintf("ReleaseSample: sample 0x%lx\n", sample));

	if (sample)
	{
		if (--sample->UseCnt == 0)
		{
			D(dprintf("ReleaseSample: free sample 0x%lx\n", sample));

			FreeMem(sample, sizeof(*sample) + sample->SampleSize);
		}
		else
		{
			D(dprintf("ReleaseSample: sample UseCnt %ld\n", sample->UseCnt));
		}
	}
}

void DeleteSample(struct Sample *sample)
{
	D(dprintf("DeleteSample: sample 0x%lx\n", sample));

	if (sample)
	{
		ReleaseSample(sample);
	}
}

#if 1

static const UBYTE ps_units[] =
{
	AHI_DEFAULT_UNIT,
	1,
	2,
	3
};
const int ps_num_units = sizeof(ps_units) / sizeof(ps_units[0]);
static struct MsgPort    ps_msgport;
static struct AHIRequest ps_ahireq;
static struct Sample    *ps_pending;

void PlaySample(struct Sample *sample, UWORD volume, UWORD period)
{
	int i;

	if (ps_msgport.mp_Node.ln_Type == 0)
	{
		ps_msgport.mp_SigBit = AllocSignal(-1);
		if ((BYTE) ps_msgport.mp_SigBit == -1)
		{
			return;
		}

		ps_msgport.mp_Node.ln_Type = NT_MSGPORT;
		ps_msgport.mp_Flags        = PA_SIGNAL;
		ps_msgport.mp_SigTask      = FindTask(NULL);
		NEWLIST(&ps_msgport.mp_MsgList);
	}

	if (ps_ahireq.ahir_Std.io_Device == NULL)
	{
		ps_ahireq.ahir_Std.io_Message.mn_Node.ln_Type = NT_REPLYMSG;
		ps_ahireq.ahir_Std.io_Message.mn_ReplyPort    = &ps_msgport;
		ps_ahireq.ahir_Std.io_Message.mn_Length       = sizeof(ps_ahireq);
		ps_ahireq.ahir_Version                        = 4;

		for (i = 0; i < ps_num_units; i++)
		{
			D(dprintf("PlaySample: Open %s unit %lu\n", AHINAME, ps_units[i]));

			if (!OpenDevice(AHINAME, ps_units[i], (APTR) &ps_ahireq, 0))
			{
				D(dprintf("PlaySample: unit %lu ok!\n", ps_units[i]));
				break;
			}
			ps_ahireq.ahir_Std.io_Device = NULL;
		}
	}

	D(dprintf("PlaySample: ps_ahireq.ahir_Std.io_Device 0x%lx\n", ps_ahireq.ahir_Std.io_Device));

	if (ps_ahireq.ahir_Std.io_Device)
	{
		ULONG freq;

		/* min freq: 3546895/3134 = 1131
		 * max freq: 3546895/73   = 48587
		 */
		freq = 3546895 / period;

		D(dprintf("PlaySample: volume %lu period %lu (freq %lu)\n",
		          volume, period, freq));

		/* Must obtain before releasing, in case it's the same
		 * sample as the pending one...
		 */
		ObtainSample(sample);

		if (ps_pending)
		{
			AbortIO((APTR) &ps_ahireq);
			WaitIO((APTR) &ps_ahireq);
			SetSignal(0, 1 << ps_msgport.mp_SigBit);

			ReleaseSample(ps_pending);
			ps_pending = NULL;
		}

		ps_ahireq.ahir_Std.io_Command = CMD_WRITE;
		ps_ahireq.ahir_Std.io_Data    = (APTR) sample->SampleData;
		ps_ahireq.ahir_Std.io_Length  = sample->SampleSize;
		ps_ahireq.ahir_Std.io_Offset  = 0;
		ps_ahireq.ahir_Type           = AHIST_M16S;
		ps_ahireq.ahir_Frequency      = freq;
		ps_ahireq.ahir_Volume         = 0x10000 * volume / 64;
		ps_ahireq.ahir_Position       = 0x8000; /* center */
		ps_ahireq.ahir_Link           = NULL;
		SendIO((APTR) &ps_ahireq);

		ps_pending = sample;
	}
}

ULONG GetWaitSampleMask(void)
{
	if (ps_pending)
	{
		return 1 << ps_msgport.mp_SigBit;
	}

	return 0;
}

void WaitSample(void)
{
	D(dprintf("WaitSample:\n"));

	if (ps_pending)
	{
		WaitIO((APTR) &ps_ahireq);
		SetSignal(0, 1 << ps_msgport.mp_SigBit);

		ReleaseSample(ps_pending);
		ps_pending = NULL;
	}
}

void StopSample(void)
{
	D(dprintf("StopSample:\n"));

	if (ps_pending)
	{
		AbortIO((APTR) &ps_ahireq);
		WaitIO((APTR) &ps_ahireq);
		SetSignal(0, 1 << ps_msgport.mp_SigBit);

		ReleaseSample(ps_pending);
		ps_pending = NULL;
	}
}

void CleanupSample(void)
{
	D(dprintf("CleanupSample:\n"));

	StopSample();

	if (ps_ahireq.ahir_Std.io_Device)
	{
		CloseDevice((APTR) &ps_ahireq);
		ps_ahireq.ahir_Std.io_Device = NULL;
	}

	if (ps_msgport.mp_Node.ln_Type == NT_MSGPORT)
	{
		FreeSignal(ps_msgport.mp_SigBit);
		ps_msgport.mp_Node.ln_Type = 0;
	}
}

#else

void PlaySample(struct Sample *sample, UWORD volume, UWORD period)
{
	struct MsgPort msgport;
	struct AHIRequest ahireq;
	static const UBYTE units[] =
	{
		AHI_DEFAULT_UNIT,
		1,
		2,
		3
	};
	const int num_units = sizeof(units) / sizeof(units[0]);
	int i;

	msgport.mp_Node.ln_Type = NT_MSGPORT;
	msgport.mp_Flags        = PA_SIGNAL;
	msgport.mp_SigBit       = AllocSignal(-1);
	if ((BYTE) msgport.mp_SigBit == -1)
	{
		msgport.mp_SigBit = SIGB_SINGLE;
	}
	msgport.mp_SigTask      = FindTask(NULL);
	NEWLIST(&msgport.mp_MsgList);

	ahireq.ahir_Std.io_Message.mn_Node.ln_Type = NT_REPLYMSG;
	ahireq.ahir_Std.io_Message.mn_ReplyPort    = &msgport;
	ahireq.ahir_Std.io_Message.mn_Length       = sizeof(ahireq);
	ahireq.ahir_Version                        = 4;

	for (i = 0; i < num_units; i++)
	{
		if (!OpenDevice(AHINAME, units[i], (APTR) &ahireq, 0))
		{
			break;
		}
	}
	if (i < num_units)
	{
		ULONG freq;
		ULONG sigs;

		/* min freq: 3546895/3134 = 1131
		 * max freq: 3546895/73   = 48587
		 */
		freq = 3546895 / period;

		D(dprintf("PlaySample: unit %lu volume %lu period %lu (freq %lu)\n",
		          units[i], volume, period, freq));

		ahireq.ahir_Std.io_Command = CMD_WRITE;
		ahireq.ahir_Std.io_Data    = (APTR) sample->SampleData;
		ahireq.ahir_Std.io_Length  = sample->SampleSize;
		ahireq.ahir_Std.io_Offset  = 0;
		ahireq.ahir_Type           = AHIST_M16S;
		ahireq.ahir_Frequency      = freq;
		ahireq.ahir_Volume         = 0x10000 * volume / 64;
		ahireq.ahir_Position       = 0x8000; /* center */
		ahireq.ahir_Link           = NULL;
		SendIO((APTR) &ahireq);

		sigs = Wait(0xffffffff);
		if (!(sigs & 1UL << msgport.mp_SigBit))
		{
			AbortIO((APTR) &ahireq);
		}
		WaitIO((APTR) &ahireq);

		CloseDevice((APTR) &ahireq);

		/* Set back any signals we got above */
		sigs &= ~(1UL << msgport.mp_SigBit);
		SetSignal(sigs, sigs);
	}

	if (msgport.mp_SigBit != SIGB_SINGLE)
	{
		FreeSignal(msgport.mp_SigBit);
	}
}

#endif
